// -*- c++ -*-
#include <stdlib.h>
#include "handle.h"
// -------------------------------------------
//  Greg Johnson and ummm me too...
//  A non-locking queue class,
//  partially stolen from "Simple, Fast, 
//  and Practical Non-Blocking and Blocking
//  Concurrent Queue Algorithms
// -------------------------------------------
// #ifndef __sgi
// //template <class T>
// bool CAS(long*v, long o, long n)
// {
//   if (*v == o){
//     *v = n;
//     return true;
//   } 
//   return false;
// }
// //#else
// //#define CAS(V,O,N) __compare_and_swap((long*)V,(long)O,(long)N)
// #endif

//#define DEBUGWFQUEUE(X) cout << "wfqueue.h - "; X 
#define DEBUGWFQUEUE(X)

template<class T> class wfqueue;

template<class T>
ostream& operator<<(ostream&, const wfqueue<T>&);

template<class T>
class wfqueue 
{
public:

 
  struct wfnode {
    T      value;
    handle<wfnode> next;
  };

  static handle<wfnode> free_list;


  typedef handle<wfnode> nodeptr;
  

  typedef   handle<wfnode>  token ;

public:
  wfqueue();
  wfqueue(const wfqueue&q);	// Deep copy
  ~wfqueue();

  // High level operatation	       
  bool peek(T&, token &tok);	// Peek in queue with token;

  bool remove(T&, const token &tok);
  bool insert(const T&, const token &tok);

  //void print_state();

  bool gettoken(token &tok) {
    DEBUGWFQUEUE(cout << "wfqueue<T>::gettoken(token &tok)" << endl);
    for (;;) {
      nodeptr tail = tail_;
      if (tail->next!=0) {
	CAS(&tail_, tail, tail->next);
      }
      else {
	break;
      }
    }
    tok = tail_;
    DEBUGWFQUEUE(cout << "wfqueue<T>::gettoken(token &tok)" << endl);
    return true;
  }

public:
  nodeptr head_;
  nodeptr tail_;
};

// template<class T>
// wfqueue<T>::nodeptr wfqueue<T>::free_list = 0;

template<class T>
wfqueue<T>::wfqueue ()
{
  DEBUGWFQUEUE(cout << "wfqueue<T>::wfqueue() - Begin" << endl);
  wfnode r;
  //nodeptr n (r);
  nodeptr n = nodeptr(r);
  n->next = 0;
  head_ = n;
  tail_ = n;
  DEBUGWFQUEUE(cout << "wfqueue<T>::wfqueue() - End" << endl);
}

template<class T>
wfqueue<T>::wfqueue (const wfqueue&q)
{
  DEBUGWFQUEUE(cout << "wfqueue<T>::wfqueue(const wfqueue&q) - Begin" << endl);
  wfnode r;
  //nodeptr n (r); 
  nodeptr n = nodeptr(r);
  n -> next = 0;
  head_ = n;
  tail_ = n;
  cerr << "wfqueue::constructor with deep copy baby - You better not be reading this.  Ever." << endl;
  for (nodeptr qn = q.head_->next; qn !=0 ; qn = qn->next) {
    n->value = qn->value;
    n->next = 0;
    if (qn->next == 0) break;
    n->next = nodeptr(r);
    n = n->next;
  }
  tail_ = n;
  DEBUGWFQUEUE(cout << "wfqueue<T>::wfqueue(const wfqueue&q) - End" << endl);
}

template<class T>
wfqueue<T>::~wfqueue ()
{
  DEBUGWFQUEUE(cout << "wfqueue<T>::~wfqueue() - Begin" << endl);
  nodeptr next;
  for (nodeptr qn = head_; qn !=0 ; qn=next) {
    next = qn->next;
  }
  DEBUGWFQUEUE(cout << "wfqueue<T>::~wfqueue() - End" << endl); 
}

// template<class T>
// void wfqueue<T>::debugger (bool values) {
//   DEBUGWFQUEUE(cout << "head = " << head_ << ", tail = " << tail_ << endl);
//   DEBUGWFQUEUE(cout << "head->next = " << head_->next << ", tail_->next = " << tail_->next << endl);
//   if (values) {
//     for (node *qn=head_; qn; qn=qn->next) {
//       DEBUGWFQUEUE(cout << "address = " << qn << " and value = " << qn->value << endl);
//     }
//   }
// }


template<class T>
bool wfqueue<T>::peek (T&v, token& tok)
{
  DEBUGWFQUEUE(cout << "bool wfqueue<T>::peek(T&v, token& tok) - Begin" << endl);
  nodeptr head;
  nodeptr tail;
  nodeptr next;

  for(;;) {
    tok = head_;
    head = head_;
    tail = tail_;
    next = head->next;
    if (head==head_) {
      if (head==tail) {
	//if head==tail we are at an empty node, OR, TAIL is screwed up.
	if (next==0) {
	  //queue is empty.
	  tok = head_;
	  DEBUGWFQUEUE(cout << "bool wfqueue<T>::peek(T&v, token& tok) - End" << endl);
	  return false;
	}
	CAS(&tail_, tail, tail_->next);
      }
      else {
	//read value before CAS, otherwise a dequeue might free the next node
	v = next->value;
	if (CAS(&tok, head_, head_)) {
	  break;
	}
    
      }
    }
  }
  DEBUGWFQUEUE(cout << "bool wfqueue<T>::peek(T&v, token& tok) - End" << endl);
  return true;
}

template<class T>
bool wfqueue<T>::remove (T&v, const token& tok)
{
  DEBUGWFQUEUE(cerr << "bool wfqueue<T>::remove(T&v,const token& tok) - Begin" << endl);
  nodeptr head;
  nodeptr tail;
  nodeptr next;

  for(;;) {
    head = head_;
    tail = tail_;
    next = head->next;
    
    //===========================================
    //inserting token code into set algorithm
    if (tok!=head) {
      DEBUGWFQUEUE(cout << "bool wfqueue<T>::remove(T&v,const token& tok) - End" << endl);
      return false;
    }
    //end of new token code 
    //===========================================

    if (head==head_) {
      if (head==tail) {
	//if head==tail we are at an empty node, OR, TAIL is screwed up.
	if (next==0) {
	  //queue is empty.
	  //v = head->value;
	  //cerr << "QUEUE IS EMPTY" << endl;
	  DEBUGWFQUEUE(cout << "bool wfqueue<T>::remove(T&v,const token& tok) - End" << endl);
	  return false;
	}
	CAS(&tail_, tail, tail_->next);
      }
      else {
	//read value before CAS, otherwise a dequeue might free the next node
	v = next->value;
	if (CAS(&head_, head, next)) {
	  break;
	}
      }
    }
  }
  //putfree(head);
  DEBUGWFQUEUE(cout << "bool wfqueue<T>::remove(T&v,const token& tok) - End" << endl);
  return true;
}

// Attempt to do a queue insert based on the token
// We do this by ensuring the tail element has not
// changed since the token was taken.d
template<class T>
bool wfqueue<T>::insert  (const T&v, const token& tok)
{
  DEBUGWFQUEUE(cout << "bool wfqueue<T>::insert(T&v,const token& tok) - Begin" << endl);
  nodeptr tail;
  nodeptr next;
  wfnode r;
  //nodeptr n (r);
  nodeptr n = nodeptr(r);
  n->value=v;
  n->next=0;
  for (;;) {
    tail = tail_;
    next = tail->next;

    //=========================================
    //start of new token code
    if (tok!=tail) {
      DEBUGWFQUEUE(cout << "bool wfqueue<T>::insert(T&v,const token& tok) - End" << endl);
      return false;
    }
    //end of new token code
    //=========================================

    if (tail==tail_) {
      if (next==0) {
	if (CAS(&tail_->next, next, n)) {
	  break;
	}
      }
      else {
	CAS(&tail_, tail, tail->next);
      }
    }
  }
  //Slight difference in algorithm, I am not gonna
  //swing directly to the new node, but rather cruise
  //the list to find it.
  for (;;) {
    tail = tail_;
    if (tail->next!=0) {
      CAS(&tail_, tail, tail->next);
    }
    else {
      break;
    }
  }
  DEBUGWFQUEUE(cout << "bool wfqueue<T>::insert(T&v,const token& tok) - End" << endl);
  return true;
}



template<class T>
ostream& operator << (ostream&o, const wfqueue<T>& q)
{
  wfqueue<T>::nodeptr n = q.head_;

  while (n) {
    o << n->value;
    n = n->next;
    if (n) o << ',';
  }
  return o;
}

