// -*- c++ -*-

#include <assert.h>
#include <string.h>

#include "btree.h"
#include <iostream.h>
#include <limits.h>

//#define NULL    (0)

#define DEBUX(X) 
//#define DEBUX(X)  X

// btree::node split_root(btree_node *root, 
// 		btree_node *newone, btree_node *left, btree_node *right);
// btree::node split_nodes(btree_node *parent, int pos,
// 			btree_node *it, btree_node *right, btree_node *left);


int btreefails = 0;

void move_operation (btree_op *op,
		     btree::op_queue *from, 
		     btree::op_queue *to, 
		     btree::op_queue::token& remtok);



ostream& operator <<(ostream&o, const btree_op&op)
{
  static char *opname[] = { "I", "S", "R", "Remove" };

  o << opname [op.operation] << '(' << op.seqnum << ')';

  switch (op.operation) {
  case btree_op::INSERT : 
    o << '[' << op.u.i.key << ']';
    break;
  case btree_op::RANGE : 
    o << '[' << op.u.r.low << ',' << op.u.r.high <<']'; break;
  }
  return o;
}

unsigned long btree_op::sequencer = 0;
  
// -------------------------------------------------
//
// -------------------------------------------------

ostream& operator<< (ostream& o,  const btree_node&n)
{
  static char*leafer[] = {"internal" , "leaf" };
  o  << leafer[n.isleaf()] << "K[";
  for (int i=0; i < n.keycount(); i++) {
    if (i != 0) o << ' ';
    o << n.key(i) ;
  }
  o << ']';
  o << "Q[" << *n.ops_ << "]";

  return o;
}

void move_operation (btree::op        op,
		     btree::op_queue *from, 
		     btree::op_queue *to, 
		     btree::op_queue::token& remtok)
{
  DEBUX(cerr << "move_operation" << endl);

  btree::op_queue::token instok;

  to->gettoken (instok);

  CAS (&op->move_token, TBOTTOM, instok);

  to->insert (btree_op (*op), op->move_token);

  from->remove (remtok);
}



// ------------------------------------------------
// btree_op
// ------------------------------------------------
btree_op::btree_op (const btree_op&op) 
  : seqnum (op.seqnum), 
    operation (op.operation), 
    move_token (TBOTTOM),	// Ensure on copy that these tokens are
    tree_token (RBOTTOM)	// reset to BOTTOM
{
  memcpy (&u, &op.u, sizeof (u));
}



// ------------------------------------------------
// btree_node
// ------------------------------------------------
btree_node::btree_node () 
  : leaf_(true), keycount_(0)
{
  ops_ = new op_queue;
  memset (&entry(0), 0, sizeof(struct node_entry) * BTREE_KEYS);
}

btree_node::~btree_node () 
{
  delete ops_;
}


bool btree_node::next_op(btree_op& op, bool all)
{
  btree_op *peek_op;

  // Figure out what is the first operation we should be doing
  // and execute that operation.  Leave it in the queue so
  // ohter processes may help if we fail.

  do { 
    wfqueue<btree_op>::token rtok;
    if (ops_->peek (peek_op, rtok) == false) return false;

    switch (peek_op->operation){
    case btree_op::INSERT : internal_insert (peek_op, rtok); 
      break;
      //    case btree_op::SEARCH : internal_search (peek_op, rtok); 
      //      break;
    case btree_op::RANGE  : internal_range (peek_op, rtok); 
      break;
    }
  } while ((*peek_op < op) || (all));

  return true;
}

btree::node btree_node::next_insert_node (btree_op& op)
{
  if (isleaf ()) {
    int v = op.u.i.key;

    int ind = scan_keys (v);

    if (val (ind) != v) {
      //cerr << "**no key=" << v << " val " << *this << endl;
      btreefails ++;
    }
    
    return 0; 
  }

  int index = scan_keys (op.u.i.key);
  return kid (index);
}

// Make a new leaf node and insert the value there..
void btree_node::insert_on_newleaf (const_node leaf, int v)
{
  assert (leaf->isleaf());
  
  int pos = scan_kids (leaf);
  assert(pos>=0);
  
  if (pos == BTREE_KEYS) {
    DEBUX (cerr << "insert_on_newleaf :: ** missing old leaf *"  << endl
	   << "    this " << *this << endl
	   << "    leaf " << *leaf << endl);

    return;
  }

  node newleaf = leaf->clone();
  newleaf->insert_node (v);
  kid (pos) = newleaf;
}



// Called when grandchild is either full or a leaf
// Return the child to be CASed
btree::node btree_node::help_insert_grand(const_node child, 
					  const_node grand,
					  int v)
{
  node newchild = child->clone();

  if (grand->isfull()) {
    // Recreate our path through the tree with new nodes (in case somone else
    // is doing the samething).

    node grandleft = new btree_node();
    node grandright= new btree_node();
    
    if (!btree_node::split_child (newchild, grand, grandleft, grandright))
      return 0;

    DEBUX (cerr << "help_insert_grand :: splitchild " << endl
	   << "         child = " << *newchild << endl
	   << "         left  = " << *grandleft << endl
	   << "         right = " << *grandright << endl);

    //int grand_index = newchild->scan_keys(v);
    if (grand->isleaf()) {
      node newleaf = newchild->find_kid (v);
      assert (newleaf == grandleft || newleaf == grandright);
      newleaf ->insert_node (v);
    }
  } else if (grand->isleaf()) {
    //}else {
    newchild->insert_on_newleaf (grand, v);

  } else {
    assert(false);
  }

  return newchild;
}

// Internal insert can only be used if the tree is three deep.
// Currently we are the gradfather node.  No other change than a 
// CAS need be done.
// ---------------------------------------------
// split the child and replace this node in parent
//          r    r'       r : Parent node
//         /    /        r' : current node (with new child)
//        c    c'        c': new node (after replacement)
//       /    / \        c : curent child node
//      g    g1 g2    g1,g2: split child nodes.
// ---------------------------------------------

bool btree_node::internal_insert(btree_op * op, op_queue::token& rtok)
{
  int k = op->u.i.val;
  int v = op->u.i.val;
  node root = this;
  

  // conditions:
  //  1.  I am at the root a tree at least 3 deep.
  //  2.  I am might be full (but my leaf is at least three deep)
  //  3.  My child is not full (would be split by upper level).
  //  4   My child may be made full by my splitting the grandchild

  // Work:
  //  1.  Empty the queue on the critical path to the insert location
  //  2.  Perform any needed splits by full nodes.
  //  3.  When grandchild is leaf, insert
  //  4.   Otherwise just move the request to the child

  int child_index = root->scan_keys (k);
  node *childptr  = &root->kid (child_index);
  node child      = root->kid (child_index);
  

  CAS (&op->tree_token, RBOTTOM, child);

  for (;;) {
    node newchild = 0;

    //child_index = root->scan_keys (k);
    //child       = root->kid (child_index);
    
    if (child -> next_op (*op)) {
      DEBUX(cerr << "internal_insert :: clearing child" << endl);
      continue;
    }

    int grand_index = child->scan_keys(k);
    node grand      = child->kid (grand_index);

    DEBUX (cerr << "internal_insert ::  "  << *op << endl
	   << "         root = " << *root << endl
	   << "         child(" << child_index<< ") = " << *child << endl
	   << "         grand(" << grand_index<< ")= " << *grand << endl);

    // Stabilize the lower nodes.
    if (grand -> next_op (*op)) { 
      DEBUX(cerr << "internal_insert :: clearing grand" << endl);
      continue;
    }

    // Decide action
    if (grand->isleaf()) {

      DEBUX(cerr << "internal_insert  :: grandleaf " << endl
	    << "         child=" << child_index << endl
	    << "         grand=" << grand_index <<endl);

      // Insert the new node swapping in the changes 
      newchild = help_insert_grand(child, grand, v);
      if (CAS (childptr, op->tree_token, newchild))
	  assert(*childptr!=0);
      ops_->remove (rtok);
      return true;

    } else if (grand->isfull ()) {
      
      // The grandkid is full -- swap in a new child and move the request
      // down (it can't be a leaf here).
      
      newchild = help_insert_grand(child, grand, v);
      
      if (CAS (childptr, op->tree_token, newchild))
	assert(*childptr!=0);

      move_operation (op, root->ops_, child->ops_, rtok);

      return true;
    }

    move_operation (op, root->ops_, child->ops_, rtok);
    return true;
  }
}

void btree_node::copyleftstuff(node tocopy) const {
 
  memcpy (&tocopy->v_[0], &v_[0], sizeof (v_[0]) * BTREE_KEYS/2);
  
}

void btree_node::copyrightstuff(node tocopy) const {
 
  memcpy (&tocopy->v_[0], &v_[BTREE_KEYS/2], sizeof (v_[0]) * BTREE_KEYS/2);
  
}

void btree_node::makesomespace(int index) {
  memmove (&entry(index+1), &entry(index), 
           (keycount_ - index)*sizeof(struct node_entry));
}

int btree_node::scan_keys (int k)  const
{
  if (keycount_==0)
    return 0;

  int nodekeys = keycount();
  int i;

  for (i=0; i < nodekeys; i++)  {
    if (k <= key(i))
      return i;
  }
  return i;
}

btree::node btree_node::find_kid (int k)  const
{
  if (keycount_==0)
    return 0;

  int nodekeys = keycount();
  int i;

  for (i=0; i < nodekeys; i++)  {
    if (k <= key(i))
      return kid(i);
  }

  return 0;
}


int btree_node::scan_kids (btree::const_node x)  const
{
  if (keycount_==0)
    return -1;

  int nodekeys = keycount();
  int i;

  for (i=0; i < nodekeys; i++)  {
    if (kid(i) == x)
      return i;
  }
  return -1;
}

// A range operation
// Rsturn the the set of nodes contained in the range
// specified by the operation in all the children of the tree.
// Arguments : ROP the operation to be completed.
//             REMTOK the removal token for this operation
//             MINE   True when this process  initiated the oepration.
//
void btree_node::internal_range(btree_op * rop, op_queue::token& remtok)
{
  int low = rop->u.r.low;
  int high = rop->u.r.high;
  btree::result_set *results = rop->u.r.results;

  btree_op::ranger_queue *tq = 0;
  //cout << "internal_range: op = " << rop << " on " << *this << endl;

  // Check to see if this operation has been completed
  if (isleaf()) {               
    // Collect the result of range query and add to result set.
    
    for (int i=0; i < keycount(); i++) {
      if (low <= key (i) && key(i) <= high) {
	//cout << "Adding " << val (i) << endl;
        results->insert (val(i));
      }
    }
    ops_->remove(remtok);

    return;
  }
  
  // We are going to try to place a set of R request on every
  // subqueue.  We do this by saving a token from each subqueue.
  // If someone has modified the queue after taking the token,
  // then the only action they could have done is to add the request.

  // Create the list of tokens if no one has

  if (rop->u.r.tokens == QBOTTOM) {
    tq = new btree_op::ranger_queue ();
    
    // Fill in the token values of the queus we need to write this op to.
    op_queue::token tok;

    for (int i=0; i < keycount(); i++) {
      if (low <= key(i)) {
        kid (i)->ops_->gettoken (tok);
        tq -> insert (btree_op::ranger (kid(i)->ops_, tok));
      }

      if (high <= key(i)) break;
    }
    
    // Attempt to switch in the token this process created.
    // The compare and swap will succeed only if I am the first.
    if (CAS (&rop->u.r.tokens, QBOTTOM, tq) == 0)
      delete tq;
  }

  // We have the list of tokens, We can insert a *copy* of Range request
  // in each the queue (those that have not been modified).
  
  btree_op::ranger *rick;
  btree_op::ranger_queue *toks = rop->u.r.tokens;
  if (toks != QTOP) {
    btree_op  r = *rop;
    r.u.r.tokens = QBOTTOM;

    btree_op::ranger_queue::token rtok;
    while (toks->peek(rick, rtok)) { 
      rick->opq->insert (r, rick->token);
      toks->remove (rtok);
    }
  }

  // All subqueues should now have a request.
  // Try to remove the token list and the request (race?)
  
  toks = rop->u.r.tokens;
  if (toks != QTOP){ 
    CAS (&rop->u.r.tokens, toks, QTOP);
    //delete tq;              // The last thread to have a copy may delete,
  }
  ops_->remove (remtok);

}


void btree_node::insert_node(int v) {
  int mykey = scan_keys(v);

  if (val (mykey) == v)  {
    DEBUX (cerr << " btree_node:: insert_node *found*" << v<<endl);  
    return;
  }

  makesomespace(mykey);
  val(mykey) = v;
  key(mykey) = v;
  keycount_++;

  DEBUX (cerr << "btree_node :: insert_node " << *this << endl);
}

// This should only be operating on a cloned parent.
bool btree_node::split_child(btree::node parent, 
				     btree::const_node oldchild,
				     btree::node leftone,
				     btree::node rightone)
{
  int pos = parent->scan_kids (oldchild);
  //hang on
  if (pos==-1) {
    return false;
  }
  int tempkey = parent->key(pos);

  oldchild->copyleftstuff(leftone);
  oldchild->copyrightstuff(rightone);
  rightone->keycount_ = BTREE_KEYS/2;
  leftone->keycount_ = BTREE_KEYS/2;

  parent->makesomespace(pos);
  parent->keycount_++;
  parent->key(pos) = leftone->key(BTREE_KEYS/2-1);
  parent->kid(pos) = leftone;
  parent->key(pos+1) = tempkey;
  parent->kid(pos+1) = rightone;


  if (oldchild->isleaf()) {
    parent->setleaf(false);
    leftone->setleaf(true);
    rightone->setleaf(true);
  }
  else {
    leftone->setleaf(false);
    rightone->setleaf(false);
    leftone->key(leftone->keycount_-1) = INT_MAX;
    //rightone->key(rightone->keycount_) = INT_MAX;
    //rightone->keycount_ ++;
  }

  return true;
}

// must be working on cloned root
btree_node *btree_node::split_root(btree::node root, 
				   btree::node leftone,
				   btree::node rightone)
{
  root->copyleftstuff(leftone);
  root->copyrightstuff(rightone);
  rightone->keycount_ = BTREE_KEYS/2;
  leftone->keycount_ = BTREE_KEYS/2;

  root->keycount_ = 2;
  root->kid(0) = leftone;
  root->kid(1) = rightone;
  root->key(0) = leftone->key((BTREE_KEYS/2)-1);
  root->key(1) = INT_MAX;

  if (root->isleaf()) {
    root->setleaf(false);
    leftone->setleaf(true);
    rightone->setleaf(true);
  } else {
    root->setleaf(false);
    leftone->setleaf(false);
    rightone->setleaf(false);

    leftone->key(leftone->keycount_-1) = INT_MAX;
    //    rightone->key(rightone->keycount_) = INT_MAX;
    //rightone->keycount_ ++;

  }
  return root;
}



// void btree_node::internal_search(btree_op & op,
// 				 op_queue::token&tok,
// 				 bool mine)
// {
// }

// ------------------------------------------------
//  BTREE
// ------------------------------------------------

btree::btree () 
{
  root_ = new btree_node ();
}


bool btree::next_op(btree_op& op, bool all)
{
  btree_op *peek_op;

  // Figure out what is the first operation we should be doing
  // and execute that operation.  Leave it in the queue so
  // ohter processes may help if we fail.

  do { 
    op_queue::token tok;

    if (ops_.peek (peek_op, tok) == false) return false;

    switch (peek_op->operation){
    case btree_op::INSERT : root_insert (peek_op, tok); break;
      //case btree_op::SEARCH : internal_search (peek_op); break;
    case btree_op::RANGE  : root_range (peek_op, tok); break;
    }
  } while ((*peek_op < op) || (all));

  return true;
}


bool btree::search (int v)
{
  //root_->search (v);
  return true;
}

// begin a range search on the tree.
// Follow the work until the range is completed.
bool btree::range (int l, int h, result_set * r)
{
  btree_op op = btree_op::range (l,h,r);

  //  Get the range into the root queue
  ops_.insert (op);
  next_op (op);

  // Should be copied to the root
  node current_node = root_;

  internal_range (op, current_node);

  return true;
}



// Move the range into the child queues 
btree::node btree::internal_range (btree_op &op, btree::node current)
{
  int low = op.u.r.low;
  int high = op.u.r.high;

  // Visit kids 
  current->next_op (op);

  for (int i=0; i < current->keycount(); i++) {
    if ( low <= current->key(i) ){
      // Do the work available until I am sure that the range request has been
      // serviced.
      internal_range (op, current->kid(i));
    }
    if ( high <= current->key(i) ) break;
  }
  return 0;
}

// Move the operation from the btree queue to the root queue...
// Many processors may be trying to do this.
bool btree::root_range (btree_op * op, op_queue::token& rtok)
{
  node root = root_;
  move_operation (op, &ops_, root->ops_, rtok);
  return true;
}


bool btree::insert (int v) 
{
  // Create a new operation 
  btree_op insert_op = btree_op::insert (v); 

  // Insert the operation in the root queue.
  // Push this operation through the nodes until it 
  // has completed.
  ops_.insert (insert_op);

  next_op (insert_op);
  node current_node = next_insert_node(insert_op);

  while (current_node) {
    current_node -> next_op(insert_op);
    current_node  = current_node -> next_insert_node (insert_op);
  }

  return true;
}

// ------------------------------------------------------
// Find the next spot we should be working on,
// if the insert is complete returns NULL.
btree::node btree::next_insert_node (btree_op & op) 
{
  // This one  pretty simple.
  int v = op.u.i.val;
  node root = root_;

  if (root->isleaf()) return 0;

  int child_index = root->scan_keys(v);
  node child = root->kid (child_index);

  if (child->isleaf()) return 0;

#if 0
  int grand_index = child->scan_keys(v);
  node grand = child -> kid (grand_index);

  if (grand->isleaf()) return 0;
#endif

  return root;
}

// ---------------------------------------------------------------
// When the root is a leaf 
// Called whenever root is full or is a leaf
btree::node btree::help_insert_root(node root, int v)
{
  node newroot = root->clone();
  if (root->isfull()) {
    node left    = new btree_node();
    node right   = new btree_node();
    btree_node::split_root (newroot, left, right);

    DEBUX (cerr << "help_insert_root :: splitroot " << endl
	   << "         root = " << *newroot << endl
	   << "         left = " << *left << endl
	   << "         right = " << *right << endl);

    if (root->isleaf ()) {
      if (v < newroot->key (0))
	left -> insert_node (v);
      else 
	right -> insert_node (v);
    }

  } else {
    assert (newroot->isleaf());
    newroot ->insert_node (v);
  }

  return newroot;
}

// ---------------------------------------------------------------
// When the root's child is a leaf 
btree::node btree::help_insert_child(node root, const_node child, int v)
{
  node newroot = root->clone();
  
  if (child->isfull()) {
    // Parent not full (case 1) child is full .. so split child 

    node childleft    = new btree_node();
    node childright   = new btree_node();
    if (!btree_node::split_child (newroot, child, childleft, childright))
      return 0;

    DEBUX (cerr << "help_insert_child :: splitchild " << endl
	   << "         root = " << *newroot << endl
	   << "         childleft = " << *childleft << endl
	   << "         childright = " << *childright << endl);
    // Insert the value into the new kid.
    if (child->isleaf ()) {
      node newleaf = newroot->find_kid (v);
      assert (newleaf == childleft || newleaf == childright);
      newleaf ->insert_node (v);
    }
   } else if (child->isleaf()) {

     newroot->insert_on_newleaf (child, v);

   } else {
     assert(false);
   }
  
  return newroot;
}


// ---------------------------------------------------------------
// When the root's grandchild is a leaf the insert is completed
// otherwise a split is done.
btree::node btree::help_insert_grand(node root, 
				     const_node child, 
				     const_node grand,
				     int v)
{
  node newroot   = root ->clone();
  node newchild  = child->clone();

  int pos = newroot -> scan_keys (v);
  if (pos==-1) {
    cout << "fuck1" << endl;
    return 0;
  }
  //assert (pos >= 0);
  newroot->kid (pos) = newchild;

  if (grand->isfull()) {
    // Recreate our path through the tree with new nodes (in case somone else
    // is doing the samething).
    
    node grandleft = new btree_node();
    node grandright= new btree_node();
    
    if (!btree_node::split_child (newchild, grand, grandleft, grandright)) {
      cout << "fuck2" << endl;
      return 0;
    }

    DEBUX (cerr << "help_insert_grand :: splitgrand " << endl
	   << "         child = " << *newchild << endl
	   << "         left = " << *grandleft << endl
	   << "         right = " << *grandright << endl);

    if (grand->isleaf()) {
      node newleaf = newchild->find_kid (v);
      assert (newleaf == grandleft || newleaf == grandright);
      newleaf ->insert_node (v);
    }
  } else if (grand->isleaf()){

    newchild->insert_on_newleaf (grand, v);

  } else {
    assert(false);
  }
  return newroot;
}

// --------------------------------------------------------------
// 
bool btree::root_insert(btree_op * op, btree_node::op_queue::token& remtok) 
{
  int v = op->u.i.val;
  node root = root_;
  node newroot = 0;

  // The first time this will succeed, thereby setting the token once
  // for this operation
  CAS (&op->tree_token, RBOTTOM, root);

  DEBUX (cerr << ">> root_insert " << *op << endl);
  for (;;) { 
    if (newroot != 0) 
      root = newroot;

    // Case 1 -- root is leaf or full
    // Make sure we are in a stable state (clear out preceding ops).
    DEBUX( cerr << "root_insert :: root = " << *root << endl);
    if (root->next_op (*op)) continue;

    if (root->isleaf()) {

      newroot = help_insert_root (root, v);

    } else if (root->isfull()) {

      newroot = help_insert_root (root, v);
      continue;

    } else {  //root is not leaf, and not full

      // case 2 -- child is leaf or full
      int  child_index = root->scan_keys (v);
      node child       = root->kid (child_index);
      
      DEBUX( cerr << "root_insert :: child  = " << *child << endl);
      
      // Make sure we are in a stable state (clear out preceding ops).
      if (child -> next_op (*op)) continue;

      if (child->isleaf()) {
	
	newroot = help_insert_child (root, child, v);
	
      } else if (child->isfull()) {

	newroot = help_insert_child (root, child, v);
	continue;

      } else {
#if 0
	// case 3 -- grandchild is leaf or full
	int   grand_index = child->scan_keys (v);
	node  grand       = child->kid (grand_index);
	
	DEBUX( cerr << "root_insert :: grand  = " << *grand << endl);

	// Make sure we are in a stable state (clear out preceding ops).
	if (grand -> next_op (*op, true)) continue;
	
 	if (grand->isleaf()) {
	  //grand is leaf
	  newroot = help_insert_grand(root, child, grand, v);
	  
	} else if (grand->isfull()) {
	  // When the grandkid is full we swap in a new tree a 
	  //grand is full
	  newroot = help_insert_grand(root, child, grand, v);

	  if (CAS (&root_, op->tree_token, newroot))
	    assert(root_!=0);
	  move_operation (op, &ops_, root_->ops_, remtok);
	  return true;

	} else {
	  // case 4 -- let the btree_node handle it
	  // Send the request to the node
	  if (newroot != 0) {	// Occurs when we arive here after an iteration
	    CAS (&root_, op->tree_token, newroot);
	  }
	  move_operation (op, &ops_, root_->ops_, remtok);
	  return true;
	}
#endif 
	// Send the request to the node
	if (newroot != 0) {	// Occurs when we arive here after an iteration
	  CAS (&root_, op->tree_token, newroot);
	}
	move_operation (op, &ops_, root_->ops_, remtok);
	return true;
      } // case 3
    } // case 2

    if (CAS (&root_, op->tree_token, newroot))
      assert(root_!=0);
    ops_.remove (remtok);
	
    // Attempt insertion of newroot
    return true;
  }
}



ostream& operator << (ostream& o, btree&b) 
{
  //b.root_->dump_level (o);

  return o;
}




#define MAKEINTERNAL(X,L,M,H) \
  (X) -> keycount_ = 3;                         \
  (X) -> setleaf (false);			\
  (X) -> key (0) = (L);				\
  (X) -> kid (0) = new btree_node();		\
  (X) -> key (1) = (M);				\
  (X) -> kid (1) = new btree_node();		\
  (X) -> key (2) = INT_MAX;			\
  (X) -> kid (2) = new btree_node();

#define MAKELEAF(X,L,M,H) \
  (X) -> keycount_ = 3;                         \
  (X) -> setleaf (true);			\
  (X) -> key (0) = (L);				\
  (X) -> val (0) = (L);				\
  (X) -> key (1) = (M);				\
  (X) -> val (1) = (M);				\
  (X) -> key (2) = (H);				\
  (X) -> val (2) = (H);

btree *btree::buildtree () 
{
  btree * b = new btree;
  // Build first layer
  btree_node *parent =   b -> root_;

  MAKEINTERNAL(parent, 300, 600, 900);

  // second layer
  MAKEINTERNAL(parent->kid(0), 100, 200, 300);
  MAKEINTERNAL(parent->kid(1), 400, 500, 600);
  MAKEINTERNAL(parent->kid(2), 700, 800, 900);

  // third layer.
  MAKELEAF(parent->kid(0)->kid(0), 1, 50, 100);
  MAKELEAF(parent->kid(0)->kid(1), 101, 150, 200);
  MAKELEAF(parent->kid(0)->kid(2), 201, 250, 300);

  MAKELEAF(parent->kid(1)->kid(0), 301, 350, 400);
  MAKELEAF(parent->kid(1)->kid(1), 401, 450, 500);
  MAKELEAF(parent->kid(1)->kid(2), 501, 550, 600);
  
  MAKELEAF(parent->kid(2)->kid(0), 601, 650, 700);
  MAKELEAF(parent->kid(2)->kid(1), 701, 750, 800);
  MAKELEAF(parent->kid(2)->kid(2), 801, 850, 900);

  return b;
}

