// -*- c++ -*-

#include <string.h>

#include "btree.h"
#include <iostream.h>
#include <limits.h>

// ------------------------------------------------
// btree_op
// ------------------------------------------------
ostream& operator <<(ostream&o, const btree_op&op)
{
  static char *opname[] = { "I", "S", "R", "Remove" };

  o << opname [op.operation] << '(' << op.seqnum << ')';

  switch (op.operation) {
  case btree_op::INSERT : 
    o << '[' << op.u.i.val << ']';
    break;
  case btree_op::RANGE : 
    o << '[' << op.u.r.low << ',' << op.u.r.high <<']'; break;
  }
  return o;
}

int btree_op::sequencer = 0;
  
// -------------------------------------------------
//
// -------------------------------------------------

ostream& operator<< (ostream& o,  btree_node&n)
{
  static char*leafer[] = {"internal" , "leaf" };
  o  << leafer[n.isleaf()] << '[';
  int keys = n.isleaf() ? n.keycount() : n.keycount() -1;
  for (int i=0; i < n.keycount(); i++) {
    if (i != 0) o << ' ';
    o << n.key(i) ;
  }
  o << ']';
  o << "Q[" << n.ops_ << "]";

  return o;
}

void btree_node::next_op(btree_op& op, bool all)
{
  btree_op peek_op;
  // Figure out what is the first operation we should be doing
  // and execute that operation.  Leave it in the queue so
  // ohter processes may help if we fail.

  do { 
    wfqueue<btree_op>::token tok;
    if (ops_.peek (peek_op, tok) == false) break;  //would mean empty queue
    switch (peek_op.operation){
    case btree_op::INSERT : internal_insert (peek_op, tok, peek_op==op); 
      break;
    case btree_op::RANGE  : internal_range (peek_op, tok, peek_op==op); 
      break;
    }
  } while ((peek_op < op) || (all));
}

void btree_node::internal_insert(btree_op & op, 
				 op_queue::token& remove_tok,
				 bool mine)
{
  cout << "INTERNAL_INSERT" << endl;
 
}

void btree_node::copyRight(node tocopy) {
  memcpy (&tocopy->v_[0], &v_[BTREE_KEYS/2], sizeof (v_[0]) * BTREE_KEYS/2);
}

void btree_node::makeSpace(int index) {
  memmove (&entry(index+1), &entry(index), (keycount_ - index)*sizeof(struct node_entry));
}

// int btree_node::scan_keys (int k) 
// {
//   if (keycount_==0)
//     return 0;

//   int nodekeys = isleaf()?keycount_:keycount_-1;
//   int i;

//   for (i=0; i < nodekeys; i++)  {
//     if (k < key(i))
//       return i;
//   }
//   return i;
// }

// void btree_node::enumerate_keys()
// {
//   cout << "keys is ";
//   for (int i = 0; i< keycount_; i++) {
//     if (isleaf()) 
//       cout << key(i) << "=" << val(i) << ",";
//     else 
//       cout << key(i) << " has kids,";
//   }
//   cout << endl;
// }

void btree_node::internal_range(btree_op & rop, 
				op_queue::token& remtok,
				bool mine)
{

  cout << "INTERNAL_RANGE" << endl;
 
}

btree::btree () 
{
  root_ = new btree_node (0);
}


void btree::next_op(btree_op& op, bool all)
{
  btree_op peek_op;

  // Figure out what is the first operation we should be doing
  // and execute that operation.  Leave it in the queue so
  // ohter processes may help if we fail.

  do { 
    op_queue::token tok;

    if (ops_.peek (peek_op, tok) == false) break;

    switch (peek_op.operation){
    case btree_op::INSERT : root_insert (peek_op, tok, peek_op==op); break;
    case btree_op::RANGE  : root_range (peek_op, tok, peek_op==op); break;
    }
  } while ((peek_op < op) || (all));
}

// begin a range search on the tree.
// Follow the work until the range is completed.
bool btree::range (int l, int h, result_set * r)
{
  btree_op range_op = btree_op::range (l, h, r);

  wfqueue<btree_op>::token tok;
  ops_.gettoken(tok);
  //again, only I proc calls this, so just loop
  while (!ops_.insert (range_op, tok)) {
    ops_.gettoken(tok);
  }

  next_op(range_op); 
}

bool btree::root_range (btree_op & op, op_queue::token rtok, bool mine)
{
  btree_op::ranger_queue::token instok, remtok;
  btree::op_queue::token root_instok, nexttok;
  op.u.r.tokens->gettoken(instok);

  root_->ops_.gettoken(root_instok);
  if (!op.u.r.tokens->peek(nexttok, remtok)) { //satisfying this would mean no token yet
    op.u.r.tokens->insert(root_instok, instok); //ok, now we have set our "work" token for this
    op.u.r.tokens->peek(nexttok, remtok);
  }
  //now, nexttok holds the token we need to insert into root
  
  root_->ops_.insert (op, nexttok);
  //ok, now remove it from the current queue with the rtok we came in with
  ops_.remove (op, rtok);
  //only call this if I am the man
  if (!mine) {
    root_->next_op (op);
  }

  return true;
}

bool btree::insert (int v) 
{
  btree_op insert_op = btree_op::insert (v); 
  insert_op.u.i.tok = 0;
  //only 1 process is calling this at a time, so just
  //loop around the insert until it gets itself in (not really wait-free)
  //(at least until it gets into this queue
  wfqueue<btree_op>::token tok;
  ops_.gettoken(tok);
  while (!ops_.insert (insert_op, tok)) {
    ops_.gettoken(tok);
  }

  next_op(insert_op);
}

bool btree::root_insert (btree_op &op, op_queue::token rtok, bool mine)
{
  cout << "ROOT_INSERT" << endl;
}


ostream& operator << (ostream& o, btree&b) 
{
 
  return o;
}
