// -*- c++ -*-

#include <strings.h>
#include <iostream.h>
#include "wfqueue.h"
#define BTREE_KEYS 4
#define QBOTTOM  ((btree_op::ranger_queue*)-1)
#define QTOP     ((btree_op::ranger_queue*)0)
#define VOLATILE

class btree_op;
class btree_node;
class btree;

// ------------------------------------------
// The Btree datastructure 
// ------------------------------------------
class btree
{
public:

  typedef wfqueue<int> result_set;
  typedef wfqueue<btree_op> op_queue;

public:

  btree ();

  bool range (int l, int h, result_set * r); 
  bool insert (int v);

  void next_op(btree_op&, bool all=false);

private:

  btree_node *root_;
  op_queue ops_;

  bool root_insert(btree_op &op, op_queue::token rtok, bool mine);
  bool root_range(btree_op &op, op_queue::token rtok, bool mine); 

  friend ostream& operator << (ostream& o, btree&b) ;

};

// ------------------------------------------
// An operation on a btree node.
// ------------------------------------------
class btree_op 
{
public:
  enum { INSERT, RANGE };

  typedef wfqueue<btree::op_queue::token> ranger_queue;

public:


  btree_op () {}

  btree_op (int op, int val) 
    : operation (op) 
    { u.i.val = (val); seqnum = INC(&sequencer);}

  btree_op (int op, int l, int h, btree::result_set * r) 
    : operation (op), results (r)
    { u.r.low=l; u.r.high=h; u.r.tokens = QBOTTOM; seqnum = INC(&sequencer);}

  static btree_op insert (int k) { 
    return btree_op(INSERT, k); 
  }
  static btree_op range (int v,int u, btree::result_set *r) { 
    return btree_op(RANGE, v, u, r); 
  }

  union {
    struct {
      int val;
      btree::op_queue::token *tok;
    } i;
    struct {
      int low, high;
      ranger_queue *tokens;
    } r;
  } u;

  static VOLATILE int sequencer;

  int operation;

  btree::result_set *results;

  int seqnum;

  bool operator == (const btree_op&op) {return (seqnum == op.seqnum);}
  bool operator != (const btree_op&op) {return (seqnum != op.seqnum);}
  bool operator < (const btree_op&op) {return (seqnum < op.seqnum);}
  bool operator <= (const btree_op&op) {return (seqnum <= op.seqnum);}

  friend ostream& operator <<(ostream&o, const btree_op&op);

};

// ------------------------------------------
// A node in the Btree
// Each node is a sequence of entries
// ------------------------------------------

struct node_entry {
  int        key;
  btree_node* kid;
};

class btree_node
{
  typedef btree_node *node;

  typedef btree::op_queue op_queue;

public:
  
  btree_node (node p) : leaf_(true), keycount_(0) 
     {memset (&entry(0), 0, sizeof(struct node_entry) * BTREE_KEYS);}

  //btree_node (const btree_node&b); 
  node  clone () { return new btree_node (*this); }

  bool isleaf () { return leaf_; } // Leaf have no children.
  void setleaf(bool v) { leaf_ = v; }
  bool isfull()  {return keycount() >= BTREE_KEYS;}

private:

  void next_op(btree_op&, bool all=false);	// Remove next op from the queue
				// and perform the operation.

  void internal_insert (btree_op&, op_queue::token& remtok, bool mine);
  void internal_range (btree_op&, op_queue::token& remtok, bool mine);

  // Access to internal stuff.
  int&  key(int index) { return v_[index].key; }
  node& kid(int index) { return v_[index].kid; }
  int&  val(int index) { return *((int*)&v_[index].kid); }
  int   keycount ()   const { return keycount_; }
  node_entry& entry (int index) { return v_[index]; }

  //insert help operations
  void makeSpace (int index);
  void copyRight (node tocopy);
  
private:
  friend class btree;
  
  op_queue ops_;
  bool    leaf_;
  int     keycount_;
  struct node_entry   v_[BTREE_KEYS+1];

  friend ostream& operator<< (ostream& o, btree_node&n);
};


