/* ---------------------------------------------------------------------
 *  tcb.c
 *  thread control block
 *  main thread header for threaded implmentation of MM
 *
 *
 * ---------------------------------------------------------------------*/
#include "tcb.h"

#define DEBUG
//#define DEBUGX(X) X
#define DEBUGX(X)

thread_control::thread_control()
{}
thread_control::~thread_control()
{
  DEBUGX(cerr << "~thread_control" << endl);
}

void thread_control::init (int threads)
{
  if (threads <= 0) threads = 1;

  cerr << "Creating " << threads << " Threads" << endl;

  pthread_mutex_init (&tcb_lock, NULL);
  pthread_cond_init (&start_cond, NULL);
  pthread_cond_init (&stop_cond, NULL);

  total_threads = threads;
  working = 0;
  threads_running = 0;

#ifdef __sgi
  cerr << "SGI Concurrency" << total_threads << endl;
  pthread_setconcurrency (total_threads);
#endif
//  pthread_attr_init (&attr);
//  pthread_attr_setschedpolicy(&attr, SCHED_FIFO);

  workers = new thread_worker[threads];

  for (int count = 0; count < threads; count++)
    {
      workers[count].rank = count;
      workers[count].TCB = this;
      if (pthread_create (&workers[count].id, NULL, //&attr, 
			  worker_thread, &workers[count]))
      {
	DEBUGX(cerr << "Problem in create\n");
	//perror ("tcb:problem in create");
	exit (0);
      }
      else
      {
	DEBUGX( cerr << "thread " << count <<" created" << endl);
	//cout << "Creating " << count << endl;
      }
      //  pthread_attr_destroy (&attr);
    }
}

/*
thread_worker thread_control::get_worker()
{
  // Get a new worker from the worker queue
  thread_worker worker;

  pthread_mutex_lock (&tcb_lock);
  while (worker_queue.empty())
    pthread_cond_wait (&stop_cond, &tcb_lock);
  
  worker = worker_queue.front(); worker_queue.pop();

  pthread_mutex_unlock (&tcb_lock);

  return worker;
}
*/

void thread_control::start_work()
{
  pthread_cond_broadcast (&start_cond);
}

void thread_control::push_task(thread_task *t)
{
  pthread_mutex_lock (&tcb_lock);
  task_queue.push_back (task_wrap(t));
  DEBUGX(cerr << "task queue " << task_queue.size() << " long" <<endl);
  pthread_mutex_unlock (&tcb_lock);
}

void thread_control::wait_empty()
{
  DEBUGX(cerr << pthread_self() << " : wait_empty" <<endl);
  pthread_mutex_lock (&tcb_lock);
  for (;;)
    {
      // Insure that the work load is empty.
      while (!task_queue.empty())
	{
	  //DEBUGX(cerr << pthread_self() <<" : wait_empty : task queue " << task_queue.size()<< endl);
	  pthread_cond_signal (&start_cond);
	  pthread_cond_wait (&stop_cond, &tcb_lock);
	}

      // If any body is still working then wait.
      bool wasworking = false;
      while (working)
	{
	  //DEBUGX(cerr << pthread_self() <<" : wait_empty : working " << working << endl);
	  pthread_cond_wait(&stop_cond, &tcb_lock);
	  wasworking = true;
	}
      if (wasworking) continue;	// possibly added to the work load.
      
      break;
    }
  pthread_mutex_unlock (&tcb_lock);
}


void thread_control::end_work()
{
  wait_empty ();
  // No work, and no load might as well clean up.
  int killed = 0;
  for (int count=0; count < total_threads; count++) {
    if (workers[count].get_status() == thread_worker::IDLE) {
      //DEBUGX(cerr << pthread_self() <<" : wait_empty : joining " <<workers[count].id << endl);
      pthread_cancel (workers[count].id);
      workers[count].status = thread_worker::DEAD;
      killed++;
    }
  }
}

/*
void thread_control::gotowork(thread_worker w)
{
  pthread_mutex_lock (&tcb_lock);

  pthread_cond_signal (&start_cond);
  pthread_mutex_unlock (&tcb_lock);

  // Force them to grab some time 
  pthread_yield();
}
*/

/////////////////////////////////////////////////////////////////////
//
thread_task::~thread_task()
{
  //cerr << "~thread_task" << endl;
}

void thread_task::action(int /*rank*/)
{}

////////////////////////////////////////////////////////////////////
//
void * worker_thread (void *arg)
{
  ((thread_worker*)arg)->worker();
  return 0;
}

void thread_worker::worker()
{
  DEBUGX(cerr << "Worker " << pthread_self() << endl);
  task_wrap w;
  //thread_task *task;
  // Worker wait forever to do more work.
  for (;;)
    {
      /* Wait on the start condition. Grab the Mutex 
	 and get what we need to do */

      pthread_mutex_lock (&TCB->tcb_lock);
      while (TCB->task_queue.empty())
	pthread_cond_wait (&TCB->start_cond, &TCB->tcb_lock);
      w = TCB->task_queue.front();       
      TCB->task_queue.pop_front();
      TCB->working++;
      //cerr << "TCB : Launch " << rank << endl;
      pthread_mutex_unlock (&TCB->tcb_lock);

      // Do the task.
      status = BUSY;

      w.ptr->action(rank);

      delete w.ptr;

      status = IDLE;

      pthread_mutex_lock (&TCB->tcb_lock);
      //cerr << "TCB : Reap " << rank << endl;
      //TCB->worker_queue.push (*this);
      TCB->working--;
      pthread_cond_signal (&TCB->stop_cond);
      pthread_mutex_unlock (&TCB->tcb_lock);
    }
}
/////////////////////////////////////////////////////////
//
void thread_barrier::init_barrier ()
{
  //init = 1;
  pthread_mutex_init (&count_lock, NULL);
  pthread_mutex_init (&gate_lock, NULL);
  pthread_cond_init (&gate_cond, NULL);
  count = 0;
  leaving = 0;
  //  DEBUGX(cerr << "init barrier :" << named << endl);
}

void thread_barrier::wait ()
{
  //  DEBUGX(cerr << "barrier::wait enter" << named << endl);

  // FIXME! This is bad and is a race condition 
  //if (init==0) init_barrier();

  pthread_mutex_lock (&gate_lock);

  while (leaving != 0)		// Wait for any leaving thread before entering
    pthread_cond_wait (&leaving_cond, &gate_lock);

  count++;
  if (count == total)		// When all have arrived 
    {				// Set the number that need to leave.
      leaving = total;
      pthread_cond_broadcast (&gate_cond); // Tell the others to go
    }

  while (count != total)	// Wait here until all have arrived
    pthread_cond_wait (&gate_cond, &gate_lock);

  leaving--;
  if (leaving == 0)		// Last one tell the early birds
    {
      count = 0;		// reset the counter
      pthread_cond_broadcast (&leaving_cond); // Open this up again.
    }
  pthread_mutex_unlock (&gate_lock);
  
  //  DEBUGX(cerr << "barrier::wait exit" << named << endl);
}

void thread_broadcast::broadcast (void *data, 
				  int bytes, 
				  int sender, 
				  int rank)
{
  //DEBUGX(cerr << "broadcast::broadcast enter" << endl);

  if (sender == rank) 
    sender_data = data;

  bar.wait();
  // Find the senders data array and copy it over.

  if (sender != rank) memcpy (data, sender_data, bytes);
  
  bar.wait();
  //DEBUGX(cerr << "broadcast::broadcast exit" << endl);
}
