// -*- c++ -*-



#include <iostream.h>
#include <iomanip.h>
#ifdef __sgi
#include <mstring.h>
#else
#include <string>
#endif

#include "btree.h"
#include "tcb.h"
// testarray[100000] = {-1};
int *testarray;
thread_control TCB;
static int VERBOSE = 0;

int static inserts = 0;
int static ranges = 0;

extern int btreefails;
int CASFAILS = 0;

inline int random_between (int low, int high)
{
  return (int)((double)random()/INT_MAX * (high-low) + low);
}

double walltime()
{
  struct timeval tv;
  gettimeofday (&tv, 0);
  return tv.tv_sec * 1E6 + tv.tv_usec;
}


static int *spots; 

static int next_num (int n) {

  static int count = 0;
  int location ;
  int first = location;

  if (count == n) return -1;

  for (;;){
    location = random_between (0, n);
    if (spots[location] == 0) {
      spots[location] = -1;
      count ++;
      return location;
    }

    //if (location == first) return -1;
    //if (location == 10000) location = 0;
  }
}

// -----------------------------------------------------------
// TASKS
// -----------------------------------------------------------

class empty : public thread_task
{
public:

  void action (int rank) {}
};


class ranger : public thread_task 
{
public :
  ranger (btree*b, int low, int high) 
    : tree_(b),low_(low), high_(high) {}

  void action (int rank) {
    btree::result_set *results = new btree::result_set;
    tree_->range (low_, high_, results);
    if (VERBOSE) {
      cout << rank << ": on range ("<<low_ <<',' << high_ << " ) found "
	   << '[' << results << ']' << endl;
      cout.flush();
    }
    
    //results->clear();
  }
private:
  btree* tree_;
  int low_, high_;
};

int myx = 0;

class inserter1 : public thread_task 
{
public:
  inserter1(btree*b, int n) : tree_ (b), total (n) {}

  void action (int rank) {
    
    //for (int x=0; x < 10; x++) {

    //int low = INC(myx);
    //int = random_between(0, 1000);
    // cout << low << endl;
    int low = next_num(total);

    tree_->insert (low);
    //}

    if (VERBOSE) { 
      cout << rank << ": inserted " << low << endl;
      cout.flush();
    }
  }

private:
  btree* tree_;
  int  total;
  int v_;
};


class both : public thread_task 
{
public :
  both (btree*b, int low, int high) 
    : tree_(b),low_(low), high_(high) {}

  void action (int rank) {

    if (random_between (0,10) < 5) {

      btree::result_set *results = new btree::result_set;
      if (VERBOSE){
	cout << 'r' << low_ << ','<<high_;
	cout.flush();
      }

      tree_->range (low_, high_, results);

      if (VERBOSE) {
	cout << rank << ": on range ("<<low_ <<',' << high_ << " ) found "
	     << '[' << results << ']' << endl;
	cout.flush();
      }
      ranges++;
      //results->clear();
    } else {
      int low = random_between(0, 1000);

      if (VERBOSE){
	cout << 'i' << low ;
	cout.flush();
      }
      tree_->insert (low);
      inserts++;
      
    }
  }
private:
  btree* tree_;
  int low_, high_;
};


static double overhead(int n) 
{
  double starttime, endtime;
  
  starttime = walltime();
  for (int c=0; c < n; c++)
    TCB.push_task (new empty ());
  
  TCB.wait_empty();

  endtime = walltime ();

  return endtime - starttime;
}


void initialize_tree (btree*btwee, int n)
{
  for (int q=0; q<n; q++) {
    int low = random_between(0, 1000);
    btwee->insert (low);
  }
  cout << "Tree initialized to " << n << " elements" << endl;
}

static char usage[] = 
"btree [-t #] [-ri][-n #][-s #]\n"
"   -t <n>  use  n threads\n"
"   -r      Do a range test\n"
"   -b      Do both range and insert\n"
"   -i      Do an insert test\n"
"   -s <n>  Make the initial tree size <n>\n"
"   -n <n>  Rounds (default=1000)\n"
"   -f <n>  Set free list <n> cells (default=1000)\n";

int free_list_sz = 0;

main(int argc, char **argv)
{
  int c;
  int threads = 1;
  int test = 'i';
  double starttime, endtime;
  long n = 1000;
  //double overh;
  string operation;
  int initial = 0;


  while ((c = getopt (argc, argv, "t:irn:vs:f:b")) != EOF)
    switch (c) {
    case 't': threads = atoi (optarg); break;
    case 'i': test = 'i'; operation = "insert"; break;
    case 'r': test = 'r'; operation = "range"; break;
    case 'b': test = 'b'; operation = "both"; break;
    case 'n': n = atoi (optarg); break;
    case 's': initial = atoi (optarg); break;
    case 'f': free_list_sz = atoi (optarg); break;
    case 'v': VERBOSE = 1; break;
    default:
      cerr << usage << endl;
      exit(0);
    }

  TCB.init (threads);

  //overh = overhead(n);

//   cout << "empty" 
//        << ":procs<" << threads << ">"
//        << ":count<" << n << ">" 
//        << ":initial tree<" << initial << ">"
//        << ":total=" << overh  << "ms" 
//        << endl;

  
  btree btwee;

  spots = new int [n];
  for (c = 0; c < n; c++) spots[c] = 0;

  if (test == 'i') {

    initialize_tree (&btwee, initial);
    
    for (int q=n; q>0; q--) {
      //TCB.push_task(new inserter (b));
      TCB.push_task (new inserter1 (&btwee, n));
    }
    
  } else if (test == 'r') {
    
    // if (initial == 0) initial = 100;

    if (VERBOSE) {
      cout << btwee;
    }

    initialize_tree (&btwee, initial);

    starttime = walltime();
    for (int x=0; x < n; x++) {
      int low = random_between(0, 999);
      int high = random_between (low+1, 1000);
      
      TCB.push_task (new ranger (&btwee, low, high));
    }      
  } else if (test == 'b') {
    
    //  if (initial == 0) initial = 100;

    if (VERBOSE) {
      cout << btwee;
    }

    initialize_tree (&btwee, initial);

    for (int x=0; x < n; x++) {
      int low = random_between(0, 999);
      int high = random_between (low+1, 1000);
      
      TCB.push_task (new both (&btwee, low, high));
    }      
  }
  starttime = walltime();
  TCB.start_work();


  TCB.end_work();

  endtime = walltime();

  
  cout << operation 
       << ":procs<" << threads << ">"
       << ":count<" << n << ">" 
       << ":initial tree<" << initial << ">"
       << ":total=" << endtime - starttime  << "us" << endl;

  if (test == 'b') {
    cout << "We did " << inserts << " inserts and " 
	 << ranges << " ranges"<<endl;
  }

  cout << "      " << (endtime - starttime )/n << "us per op" << endl;

  cout << "verify values:" << endl;

  testarray = new int [n];

  for (int gggg=1; gggg<n;gggg++) {
    testarray[gggg] = -1;
  }
  //testarray = new int[1000];
  btwee.print_tree(testarray);
  int totalmissing =0;
  for (int ggg=1; ggg<n;ggg++) {
    if (testarray[ggg]!=ggg)
      // cout << "Position " << ggg << " is not " << ggg << " it is " << testarray[ggg] << endl;
      totalmissing++;
  }
//cout << ggg << ",";
    cout << "Total Lost:" << totalmissing << endl;
  
  exit(0);
}



// void simple_btree () 
// {
//   btree b;
//   btree::result_set r;
//   int k;

//   for ( k=0; k < 200 ; k ++) {
//     if (b.insert (k) == false) { 
//       cout << "insert failed on " << k << endl;
//     }   
//   }

//   for ( k=0; k < 200 ; k ++) {
//     if (b.search (k) == false) 
//       cout << "search failed on " << k << endl;
//   }

//   if (b.range (0, 200, &r)==false) {
//       cout << "range failed on " <<  endl;
//   }

//   if (r.count () != 200) {
//     int k;
//     int spot = 0;
//     cout << "Values in result set :";
//     while (r.count()) {
//       r.remove (k);
//       cout << setw (4) << k;
//       if (++spot % 20 == 0) cout << endl;
//     }
//     cout << endl;
//   }
// }

void reallysimpletree() {
  btree b;
  
  for (int k=20; k >0 ; k --) {
    if (b.insert (k) == false) { 
      cout << "insert failed on " << k << endl;
    }   
  }
}  
