// -*- c++ -*-
#include <stdlib.h>
// -------------------------------------------
//  Greg Johnson and ummm me too...
//  A non-locking queue class,
//  partially stolen from "Simple, Fast, 
//  and Practical Non-Blocking and Blocking
//  Concurrent Queue Algorithms
// -------------------------------------------
// #ifndef __sgi
// #define CAS(V,O,N) ((*(long*)(V)==(long)(O))?(*(long*)(V)=(long)(N),1) : 0)
// #else
// #define CAS(V,O,N) __compare_and_swap((long*)V,(long)O,(long)N)
// #endif

extern int CASFAILS;

#ifndef __sgi
template <class T>
bool CAS(T*v, const T &o, const T &n)
{
  if (*v == o){
    *v = n;
    return true;
  } 
  CASFAILS++;
  return false;
}
#else 
template <class T>
 bool CAS(T*v, const T& o, T n)
{
  bool r = __compare_and_swap((long*)v,(long)o,(long)n);
  if (!r)  CASFAILS++;
  return r;
}
#endif

#define VOLATILE 
#define FREELIST 1024*1024*32/sizeof(wfqueue<T>::node)


template<class T>
inline T INC(T&val) {
  T valAct = val;
  while (!CAS(&val, valAct, valAct+1))
    valAct = val;
  return valAct;
}

template<class T>
inline T DEC(T&val) {
  T valAct = val;
  while (!CAS(&val, valAct, valAct+1))
    valAct = val;
  return valAct;
}


extern int free_list_sz;
template<class T> class wfqueue;

template<class T>
ostream& operator<<(ostream&, const wfqueue<T>&);

template<class T>
class wfqueue 
{
public:
  struct node {
    node * next;
    T      value;
  };

  typedef   node*  token ;

public:
  wfqueue();
  wfqueue(const wfqueue&q);	// Deep copy
  ~wfqueue();

  // High level operatation	       
  bool peek(T* &, token &tok);	// Peek in queue with token;

  bool remove(const token &tok);
  bool insert(const T&, const token &tok);
  bool insert(const T&);

  void debugger(bool values=false);

  bool gettoken(token &tok) {
    for (;;) {
      node *tail = tail_;
      if (tail->next!=0) {
	CAS(&tail_, tail, tail->next);
      }
      else {
	break;
      }
    }
    tok = tail_;
    return true;
  }


private:
  static node VOLATILE* free_list;

  static node* nextfree () { 
    node VOLATILE *t;
    if (free_list == 0) {
      cerr << "Free list exhausted" << endl;
      exit (0);
    }
    for (;;) {
      t = free_list; 
      if (CAS(&free_list, t, t->next))
	break;
    }
    return t;
  }

  static void putfree (node*p) { 
    node *list;
    for (;;) {
      list = free_list;
      p->next = list; 
      if (CAS (&free_list, list, p))
	break;
    }    
  }

  static void init_freelist() {
    init_free(FREELIST);
  }

  static void init_freelist(int free_size) {
    init_free(free_size);
  }
  
  static void init_free(int free_size) {
    free_list_sz = free_size;
    //if (free_list_sz == 0) free_list_sz = FREELIST;
    node *n =  new node [free_list_sz];
    free_list = n;
    for (int x=1; x < free_list_sz; x++){
      n-> next = &free_list[x];
      n  = n ->next;
    }
    n -> next = 0;
  }

public:
  VOLATILE node *head_;
  VOLATILE node *tail_;
};

template<class T>
VOLATILE wfqueue<T>::node * wfqueue<T>::free_list = 0;

template<class T>
wfqueue<T>::wfqueue ()
{
  if (free_list == 0) 
    init_freelist();

  node* n = nextfree(); 
  n->next = 0;
  head_ = tail_ = n;
}

template<class T>
wfqueue<T>::wfqueue (const wfqueue&q)
{
  // Create an empty Queue
  node* n = nextfree(); 
  n->next = 0;
  head_ = tail_ = n;

#if 0
  node* n = nextfree(); 
  n -> next = 0;
  head_ = n;
  tail_ = n;
  cerr << "wfqueue::constructor with deep copy baby - You better not be reading this.  Ever." << endl;
  for (node *qn = q.head_->next; qn; qn = qn->next) {
    n->value = qn->value;
    n->next = 0;
    if (qn->next == 0) break;
    n->next = nextfree();
    n = n->next;
  }
  tail_ = n;
#endif
}

template<class T>
wfqueue<T>::~wfqueue ()
{
  cerr << "~wfqueue - You really think that this is safe?" << endl;
  node *next;
  for (node *qn = head_; qn; qn=next) {
    next = qn->next;
    putfree (qn);
  }
}



template<class T>
void wfqueue<T>::debugger (bool values) {
  cout << "head = " << head_ << ", tail = " << tail_ << endl;
  cout << "head->next = " << head_->next << ", tail_->next = " << tail_->next << endl;
  if (values) {
    for (node *qn=head_; qn; qn=qn->next) {
      cout << "address = " << qn << " and value = " << qn->value << endl;
    }
  }
}


template<class T>
bool wfqueue<T>::peek (T* &v, token& tok)
{

  node *head;
  node *tail;
  node *next;

  for(;;) {
    tok = head_;
    head = head_;
    tail = tail_;
    next = head->next;
    if (head==head_) {
      if (head==tail) {
	//if head==tail we are at an empty node, OR, TAIL is screwed up.
	if (next==0) {
	  //queue is empty.
	  tok = head_;
	  return false;
	}
	CAS(&tail_, tail, tail_->next);
      }
      else {
	//read value before CAS, otherwise a dequeue might free the next node
	v = &next->value;
	if (CAS(&tok, head_, head_)) {
	  break;
	}
    
      }
    }
  }
  return true;
}

template<class T>
bool wfqueue<T>::remove (const token& tok)
{

  node *head;
  node *tail;
  node *next;

  for(;;) {
    head = head_;
    tail = tail_;
    next = head->next;
    
    //===========================================
    //inserting token code into set algorithm
    if (tok!=head) {
      return false;
    }
    //end of new token code 
    //===========================================

    if (head==head_) {
      if (head==tail) {
	//if head==tail we are at an empty node, OR, TAIL is screwed up.
	if (next==0) {
	  //queue is empty.
	  return false;
	}
	CAS(&tail_, tail, tail_->next);
      }
      else {
	//read value before CAS, otherwise a dequeue might free the next node
	if (CAS(&head_, head, next)) {
	  break;
	}
      }
    }
  }
  //putfree(head);
  return true;
}

// Attempt to do a queue insert based on the token
// We do this by ensuring the tail element has not
// changed since the token was taken.d
template<class T>
bool wfqueue<T>::insert  (const T&v, const token& tok)
{
  node* tail;
  node* next;
  node* n = nextfree();
  n->value=v;
  n->next=0;
  for (;;) {
    tail = tail_;
    next = tail->next;

    //=========================================
    //start of new token code
    if (tok!=tail) {
      return false;
    }
    //end of new token code
    //=========================================

    if (tail==tail_) {
      if (next==0) {
	if (CAS(&tail_->next, next, n)) {
	  break;
	}
      }
      else {
	CAS(&tail_, tail, tail->next);
      }
    }
  }
  //Slight difference in algorithm, I am not gonna
  //swing directly to the new node, but rather cruise
  //the list to find it.
  for (;;) {
    tail = tail_;
    if (tail->next!=0) {
      CAS(&tail_, tail, tail->next);
    }
    else {
      break;
    }
  }
  return true;
}




template<class T>
bool wfqueue<T>::insert  (const T&v)
{
  //cerr << "wfqueue::insert, no token.  We don't like you anymore." << endl;
  node* tail;
  node* next;
  node* n = nextfree();
  n->value=v;
  n->next=0;

  for (;;) {
    tail = tail_;
    next = tail->next;

    if (tail==tail_) {
      if (next==0) { 
	if (CAS(&tail_->next, next, n)) {
	  break;
	}
      }
      else {
	CAS(&tail_, tail, tail->next);
      }
    }
  }

  //Slight difference in algorithm, I am not gonna
  //swing directly to the new node, but rather cruise
  //the list to find it.

  for (;;) {
    tail = tail_;
    if (tail->next!=0) {
      CAS(&tail_, tail, tail->next);
      continue;
    }
    break;
  }
  return true;
}


template<class T>
ostream& operator << (ostream&o, const wfqueue<T>& q)
{
  wfqueue<T>::node *n = q.head_->next;

  while (n) {
    o << n->value;
    n = n->next;
    if (n) o << ',';
  }
  return o;
}

